#!/usr/local/bin/perl
# ------------------------------------------------------------------------ 
# TCP Port Scanner v2.2      Copyright (C) 1998 Dave Edis <dave@edis.org>
# ------------------------------------------------------------------------ 
# Connect to a host/ip and scan selected TCP ports or all ports 1-1024
# and report whether a connection can be made on each port.
#
# This program is distributed under the terms of the GNU General Public
# License Version 2 as published by the Free Software Foundation.
# For the latest version and licence info go to http://www.edis.org/perl/
# ------------------------------------------------------------------------ 
# v1.0 Created: 29.Nov.97 - Created by Dave Edis <dave@edis.org>
# v1.1 Revised: 17.Dec.97 - Error Routine and more port descriptions
# v2.0 Revised: 20.Dec.97 - Connect Timeout, new interface, etc
# v2.1 Revised: 21.Dec.97 - Minor Interface Change, Fixed Windows problem
# v2.2 Revised: 03.Jan.98 - Display Port/Service Text Banner

$ver      = "v2.2";
$ver_date = "03.JAN.98";

# ------------------------------------------------------------------------ 
# Variables : Define variables used in program
# ------------------------------------------------------------------------ 

&Get_Input;                               # Read user form input
&Port_List;                               # Load Port Descriptions
$cgidir   = ($0=~m#^(.*)(\\|/)#,$1);      # full dir of script
$logfile  = "$cgidir/portscan.log";       # logfile
$logging  = 0;     # 1 = on 0 = off       # log scans to log file
$|++;                                     # Turn off buffering

# Common ports > 1024 to scan
@odd_ports = (2000,2121,3000,4000,5000,6000,7000,8000,8080,8181,9000);

$timeout      = 1;                       # Connect Timeout after # seconds
$SIG{ALRM}    = sub { return "Timed Out" };
$SIG{__DIE__} = \&HTML_Error;         # on die show error HTML

use Socket;
$PROTO = getprotobyname('tcp')||0; # Get TCP Protocol or use default

# ------------------------------------------------------------------------ 
# Main : Test conditions and give commands
# ------------------------------------------------------------------------ 

  if ($host) { &Scan;      }  
  else       { print &Template("$cgidir/portscan_help.html",'html'); }

exit;

# ------------------------------------------------------------------------ 
# Get Input : Reads user form input into variables or hash array
# ------------------------------------------------------------------------ 

sub Get_Input {

  my($name, $value, $pair, @pairs, $buffer);

  if ($ENV{'REQUEST_METHOD'} eq 'POST') { read(STDIN, $buffer, $ENV{'CONTENT_LENGTH'}); }
  if ($ENV{'REQUEST_METHOD'} eq 'GET')  { $buffer = $ENV{'QUERY_STRING'}; }

  @pairs = split(/&/, $buffer);
  foreach $pair (@pairs) {
    ($name, $value) = split(/=/, $pair);
    $value =~ tr/+/ /;
    $value =~ s/%([A-F0-9]{2})/pack("C", hex($1))/egi;
    ${$name} = $value;
    }
  foreach (@ARGV) { ${$_}=1; }
  }

# ------------------------------------------------------------------------ 
# Port List : Create array of know port services
# ------------------------------------------------------------------------ 

sub Port_List {

  # If we have a unix port/service listing read in port/service descriptions
    if (-e "/etc/services") {
      open(FILE,"/etc/services");
      foreach(<FILE>) { if (m|(\w+)\s+(\d+)/tcp|) { $PORT[$2]=$1; }}
      close(FILE);
      }

  $PORT[7]   = "Echo        "; $PORT[9]   = "Discard     "; $PORT[13]  = "Daytime     ";
  $PORT[17]  = "QOTD        "; $PORT[19]  = "Chargen     "; $PORT[21]  = "FTP         ";
  $PORT[22]  = "SSH         "; $PORT[23]  = "Telnet      "; $PORT[25]  = "SMTP        ";
  $PORT[37]  = "Time        "; $PORT[43]  = "Whois       "; $PORT[53]  = "DNS         ";
  $PORT[70]  = "Gopher      "; $PORT[79]  = "Finger      "; $PORT[80]  = "HTTP        ";
  $PORT[106] = "Poppassd    "; $PORT[109] = "POP2        "; $PORT[110] = "POP3        ";
  $PORT[113] = "Ident       "; $PORT[119] = "NNTP        "; $PORT[135] = "Loc-svr     ";
  $PORT[137] = "Netbios-ns  "; $PORT[138] = "Netbios-dgm "; $PORT[139] = "Netbios-ssn ";
  $PORT[143] = "IMAP2       "; $PORT[220] = "IMAP3       "; $PORT[389] = "LDAP        ";
  $PORT[443] = "HTTPS       ";

  }

# ------------------------------------------------------------------------ 
# Scan : Run through a list of ports and try to connect to each one
# ------------------------------------------------------------------------ 

sub Scan {
 
  if ($host =~ /[a-z]$/) { $paddr = inet_aton("$host.") || die("Couldn't resolve '$host'"); }
  else                   { $paddr = inet_aton("$host") || die("Couldn't resolve '$host'"); }
  ($name,$aliases,$addrtype,$length,@addrs) = gethostbyaddr($paddr,AF_INET);
  ($a,$b,$c,$d)=unpack("C4",$paddr); $ip  = "$a.$b.$c.$d";
  foreach (@addrs) { ($a,$b,$c,$d)=unpack("C4",$_); $ip2 .= "$a.$b.$c.$d "; }
 
  foreach ('all','odd') { if (${$_}) { ${"$_"."_checked"} = "checked"; }}
  &Log("Scanning $host $ip $name $aliases $ip2");
  print &Template("$cgidir/portscan.html",'html');

  if ($all) {
    for $port (1..1024) {
      ($status,$banner) = &Connect($port);
      &HTML_Scan_Port;
      $ports_scanned++;
      }
    }

  if ($odd) {
    for $port (@odd_ports) {
      ($status,$banner) = &Connect($port);
      &HTML_Scan_Port;
      $ports_scanned++;
      }
    }

  unless ($ports_scanned) { $ports_scanned = "0"; }
  $exectime = time-$^T; 
  $execmin = int($exectime/60);
  $execsec = $exectime%60;
  print qq|<br><font face="arial" size=2>Scanned $ports_scanned ports on $host in $execmin min $execsec seconds</font>|;
  }


# ------------------------------------------------------------------------ 
# Connect : Attempt to connect to a TCP port/service and return 1 or 0
#           usage : &Connect($port);
# ------------------------------------------------------------------------ 

sub Connect {

  my($status,$banner);

  alarm $timeout; 

  socket(SOCK,AF_INET,SOCK_STREAM,$PROTO)          || return "No Socket $!";
  connect(SOCK,pack "SnA4x8",AF_INET,$_[0],$paddr) || return "No Connect $PORT[$port]";
  $banner = <SOCK>;
  close(SOCK);

  $status = qq|<font color="#FFFFFF">Connected $PORT[$port]</font>|;
  $banner =~ s/\s+/ /gs;

  return ($status,$banner);
  }

# ------------------------------------------------------------------------ 
# Log : Make an entry in the log file if logging is enabled
# ------------------------------------------------------------------------ 

sub Log {

  if ($logging) {
    $date = localtime(time); 
    open (FILE,">>$logfile") || die("Can't open $logfile");
    print FILE "$date - $_[0] - $ENV{'REMOTE_HOST'} ($ENV{'REMOTE_ADDR'})\n";
    close(FILE);
    }
}

# ------------------------------------------------------------------------ 
# HTML : HTML screens used in the program
# ------------------------------------------------------------------------ 

sub HTML_Scan_Port {
  print <<HTML;
<table border=0 width=580 cellspacing=0 cellpadding=0><tr>
<td valign=top><font face="arial" size=1><b>$ip:$port : $status</b></font></td>
<td width=400 align=right><font face="arial" size=1 color="#FFFFFF"><b>$banner</b></font></td>
</tr></table>

HTML
}

sub HTML_Error {
  print "Content-type: text/html\n\n" unless ($ContentType++ > 0);
  print "ERROR: $_[0]";
  exit;
  }


# ---------------------------------------------------------------------------- 
# Template : Open a template, translate special vars, and return contents
#            usage: $var = &Template("filepath");
#                   print &FILE("filepath",'html'); for html content header
#                   print &FILE("filepath",'text'); for text content header
# ---------------------------------------------------------------------------- 

sub Template {  

  local(*FILE);

  if    ($_[1] eq 'html') { print "Content-type: text/html\n\n"  unless ($ContentType++ > 0); }
  elsif ($_[1] eq 'text') { print "Content-type: text/plain\n\n" unless ($ContentType++ > 0); }

  if    (!$_[0])          { return "Template : No file was specified<br>\n"; }
  elsif (-f "$_[0]") {
    open(FILE, "<$_[0]") || return "Template : Couldn't open $_[0]<br>\n";
    while (<FILE>) {
      s/<!-- insert="(.*?)" -->/\1/g;                     # remove <!-- insert="" -->
      s/\$([a-z0-9-_]+)\$/${$1}/gi;                       # matches $scalars$
      s/@([a-z0-9-_]+)@/@{$1}/gi;                         # matches @arrays@
      s/\$([a-z0-9-_]+)\[(\d+?)\]\$/${$1}[$2]/gi;         # matches $array[val]$
      s/\$([a-z0-9-_]+){'([a-z0-9-_]+)'}\$/$1->{$2}/gi;   # matches $hash{'val'}$\
      $FILE .= $_;
      }
    close(FILE);
    } else { return "Template : File '$_[0]' does not exist<br>\n"; }

  return $FILE;
  }

# ------------------------------------------------------------------------ 
#                                 Programming by Dave Edis <dave@edis.org>
